/* pathnames.h -- absolute filenames that bash wants for various defaults. */

/* Copyright (C) 1987,1991 Free Software Foundation, Inc.

   This file is part of GNU Bash, the Bourne Again SHell.

   Bash is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   Bash is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with Bash; see the file COPYING.  If not, write to the Free
   Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#if !defined (_PATHNAMES_H_)
#define _PATHNAMES_H_

/* The default file for hostname completion. */
#define DEFAULT_HOSTS_FILE "/etc/hosts"

/* The default login shell startup file. */
#define SYS_PROFILE "/etc/profile"

#endif /* _PATHNAMES_H */
